Package: dashboardapi
Title: Access Japan's Statistics Dashboard API
Version: 0.1.0
Authors@R: 
    person("Kenji", "Miyazaki", email = "kenjimyzk@gmail.com",
           role = c("aut", "cre"))
Description: An unofficial, tidy, rate-conscious interface to the Statistics
    Dashboard Web API provided by the Statistics Bureau of Japan. Search
    indicators, regions, terms, social events, and statistical surveys, then
    retrieve observations as normalized long or wide data frames. The API
    requires no registration or API key.
URL: https://github.com/kenjimyzk/dashboardapi
BugReports: https://github.com/kenjimyzk/dashboardapi/issues
License: MIT + file LICENSE
Depends: R (>= 4.1.0)
Imports: httr2 (>= 1.1.0), tibble
Suggests: knitr, rmarkdown, testthat (>= 3.2.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-31 14:37:16 UTC; kenjimyzk
Author: Kenji Miyazaki [aut, cre]
Maintainer: Kenji Miyazaki <kenjimyzk@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-08 11:10:02 UTC

# Generated by roxygen2: do not edit by hand

export(dashboard_api_credit)
export(dashboard_codes)
export(dashboard_data)
export(dashboard_events)
export(dashboard_indicators)
export(dashboard_regions)
export(dashboard_search)
export(dashboard_surveys)
export(dashboard_terms)

English | [日本語](https://github.com/kenjimyzk/dashboardapi/blob/main/README.ja.md)

# dashboardapi

[![R-CMD-check](https://github.com/kenjimyzk/dashboardapi/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/kenjimyzk/dashboardapi/actions/workflows/R-CMD-check.yaml)

`dashboardapi` is an unofficial R package for accessing the Statistics
Dashboard Web API provided by the Statistics Bureau of Japan. The API requires
no registration or API key.

The main workflow follows the same idea as `bojapi` and `WDI`: search metadata
for an indicator code, find any needed region codes, and retrieve observations
as a tidy data frame.

## Features

- `dashboard_search()` / `dashboard_indicators()`: find indicator codes and
  inspect cycle, region, seasonal-adjustment, unit, source, and coverage
- `dashboard_regions()`: find Japanese municipality or ISO country codes
- `dashboard_data()`: retrieve normalized long or wide observations
- `dashboard_terms()`, `dashboard_events()`, and `dashboard_surveys()`: access
  the API's auxiliary metadata
- `dashboard_codes()`: inspect common API enumerations offline
- Automatic batching beyond the API limits of five indicators and 50 regions
- English and Japanese API responses, retries, timeouts, and rate-conscious
  pauses between automatically batched requests
- Typed API, HTTP, parsing, and no-data conditions

## Installation

Install the development version from GitHub:

```r
install.packages("remotes")
remotes::install_github("kenjimyzk/dashboardapi")
```

To install from a local clone of the repository:

```r
remotes::install_local(".")
```

## Quick start

```r
library(dashboardapi)

# 1. Find an indicator
population_meta <- dashboard_search(
  "Total population (Both sexes)",
  lang = "en"
)

population_meta[, c(
  "indicator_code", "name", "cycle_name",
  "regional_rank_name", "unit"
)]

# 2. Inspect region codes
prefectures <- dashboard_regions(
  parent_region_code = "00000",
  lang = "en"
)

# 3. Retrieve annual Japanese population
population <- dashboard_data(
  indicator_code = c(population = "0201010000000010000"),
  region_code = "00000",
  time_from = "2020CY00",
  time_to = "2024CY00",
  cycle = "year",
  regional_rank = "japan",
  seasonal = "original",
  lang = "en"
)
```

The `time` column preserves the API's exact period code. The `date` column
contains the first day of the represented period for analysis:
`20240100` becomes 2024-01-01, `20242Q00` becomes 2024-04-01,
`2024CY00` becomes 2024-01-01, and `2024FY00` becomes 2024-04-01.

## Prefecture example

```r
prefecture_population <- dashboard_data(
  indicator_code = c(population = "0201010000000010000"),
  region_code = c("13000", "27000"),
  time_from = "2020CY00",
  time_to = "2024CY00",
  cycle = "year",
  regional_rank = "prefecture",
  seasonal = "original",
  wide = TRUE
)
```

## Request rate and errors

The official API page asks users not to generate a large access volume in a
short period. A request with more than five indicators or 50 regions is split
automatically, and `dashboardapi` waits at least one second between those
requests.

```r
options(
  dashboardapi.wait = 2,
  dashboardapi.timeout = 60,
  dashboardapi.retries = 3
)
```

API errors have class `dashboard_api_response_error`, communication errors
have class `dashboard_http_error`, and unexpected structures have class
`dashboard_parse_error`. A successful no-data response issues a
`dashboard_no_data_warning` and returns a typed empty tibble.

## Attribution and credit for public services

The official Copyright Policy distinguishes three requirements when content
from the Statistics Dashboard is published:

1. Cite the source.
2. If content or data are edited, say so and identify the editing entity.
3. If a service uses the API, display the specified API credit.

`dashboard_api_credit()` returns the corresponding source citation,
edited-content template, and API credit. Replace `<entity>` before publishing
edited content:

```r
attribution <- dashboard_api_credit("en")
attribution$source
attribution$processed
attribution$credit
```

`dashboardapi` is independently developed and is not affiliated with or
endorsed by the Statistics Bureau of Japan. Its MIT license does not relicense
government or third-party data, metadata, code systems, or official texts.
Check any individual rights statement before reusing third-party content.

## Official documentation

- [Statistics Dashboard API](https://dashboard.e-stat.go.jp/en/static/api)
- [Statistics Dashboard site policy](https://dashboard.e-stat.go.jp/en/static/sitePolicy)

#' dashboardapi: Japan's Statistics Dashboard data for R
#'
#' `dashboardapi` is an unofficial, independently developed client for the
#' Statistics Dashboard Web API provided by the Statistics Bureau of Japan.
#' It is not affiliated with, endorsed by, or maintained by the Statistics
#' Bureau. A typical workflow is:
#'
#' 1. Find a series with [dashboard_search()].
#' 2. Find region codes with [dashboard_regions()].
#' 3. Retrieve observations with [dashboard_data()].
#'
#' No registration or API key is required. The API documentation asks clients
#' not to create a large volume of access in a short period. When
#' [dashboard_data()] must split a request into multiple API calls, the package
#' therefore waits at least one second between calls.
#'
#' @section Package options:
#' - `dashboardapi.lang`: default response language, `"en"` or `"jp"`.
#' - `dashboardapi.wait`: seconds between automatically batched requests;
#'   default and enforced minimum `1`.
#' - `dashboardapi.timeout`: request timeout in seconds; default `30`.
#' - `dashboardapi.retries`: retries for transient failures; default `3`.
#'
#' @section Terms of use:
#' When publishing content retrieved with this package, cite the Statistics
#' Dashboard as the source. When publishing edited content, also identify the
#' editing entity. Services published using the API should show the credit
#' requested by the Statistics Dashboard. Call [dashboard_api_credit()] for
#' the source citation, edited-content template, API credit, and official
#' links. The package's MIT license applies only to original package code and
#' documentation; it does not relicense government or third-party data,
#' metadata, or official texts. Check any individual rights statement before
#' reusing third-party content.
#'
#' @seealso
#' [Statistics Dashboard API](https://dashboard.e-stat.go.jp/en/static/api)
#'
#' @keywords internal
"_PACKAGE"

.dashboardapi_base_url <- "https://dashboard.e-stat.go.jp/api/1.0/Json"
.dashboardapi_max_indicators <- 5L
.dashboardapi_max_regions <- 50L

#' Return commonly used Statistics Dashboard API codes
#'
#' The API's full category, survey, and region code systems are maintained by
#' the Statistics Bureau. This helper returns the small enumerations used
#' directly by package arguments.
#'
#' @param type One of `"cycle"`, `"regional_rank"`, `"seasonal"`,
#'   `"provisional"`, or `"social_event_level"`.
#' @param lang Label language, `"en"` or `"jp"`.
#'
#' @return A tibble containing `code`, `name`, and `type`.
#' @export
#' @examples
#' dashboard_codes("cycle")
#' dashboard_codes("regional_rank", lang = "jp")
dashboard_codes <- function(
    type = c(
      "cycle", "regional_rank", "seasonal",
      "provisional", "social_event_level"
    ),
    lang = c("en", "jp")) {
  type <- match.arg(type)
  lang <- match.arg(lang)
  values <- switch(
    type,
    cycle = list(
      code = as.character(1:4),
      en = c("Month", "Quarter", "Calendar year", "Fiscal year"),
      jp = c(
        "\u6708", "\u56DB\u534A\u671F",
        "\u5E74", "\u5E74\u5EA6"
      )
    ),
    regional_rank = list(
      code = as.character(1:4),
      en = c("Country", "Nationwide (Japan)", "Prefecture", "Municipality"),
      jp = c(
        "\u56FD", "\u5168\u56FD\uFF08\u65E5\u672C\uFF09",
        "\u90FD\u9053\u5E9C\u770C", "\u5E02\u533A\u753A\u6751"
      )
    ),
    seasonal = list(
      code = c("1", "2"),
      en = c("Original series", "Seasonally adjusted series"),
      jp = c(
        "\u539F\u6570\u5024",
        "\u5B63\u7BC0\u8ABF\u6574\u5024"
      )
    ),
    provisional = list(
      code = c("0", "1"),
      en = c("Fixed", "Preliminary"),
      jp = c("\u78BA\u5831", "\u901F\u5831")
    ),
    social_event_level = list(
      code = as.character(1:3),
      en = c("Low", "Medium", "High"),
      jp = c("\u4F4E", "\u4E2D", "\u9AD8")
    )
  )
  tibble::tibble(
    code = values$code,
    name = values[[lang]],
    type = type
  )
}

fixture_indicator_root <- function(status = "0") {
  list(
    RESULT = list(status = status, errorMsg = "Success."),
    METADATA_INF = list(
      CLASS_INF = list(
        CLASS_OBJ = list(
          list(
            "@name" = "Total population (Both sexes)",
            "@code" = "0201010000000010000",
            details = list(
              detail = list(
                list(
                  "@code" = "0201010001",
                  "@name" = "Total population",
                  "$" = "Population definition."
                )
              )
            ),
            annotations = list(
              list(
                "@cycle" = "3",
                "@regionalRank" = "2",
                "@isSeasonal" = "1",
                "@annotation" = "Sample annotation."
              )
            ),
            CLASS = list(
              list(
                "@name" = "Total population (Both sexes)",
                "@sname" = "Total population",
                "@fromDate" = "1920CY00",
                "@toDate" = "9999CY00",
                cycle = list("@code" = "3", "@name" = "Calendar Year"),
                RegionalRank = list("@code" = "2", "@name" = "Japan"),
                IsSeasonal = list("@code" = "1", "@name" = "Original Series"),
                "@statName" = "Population Census / Population Estimates",
                "@unit" = "person"
              )
            )
          )
        )
      )
    )
  )
}

fixture_region_root <- function() {
  list(
    RESULT = list(status = "0", errorMsg = "Success."),
    METADATA_INF = list(
      CLASS_INF = list(
        CLASS_OBJ = list(
          list(
            "@parentRegionCode" = "00000",
            "@name" = "Japan",
            "@hiragana" = "",
            CLASS = list(
              list(
                "@regionCode" = "13000",
                "@name" = "Tokyo-to",
                "@level" = "3",
                "@hiragana" = "",
                "@fromDate" = "190001",
                "@toDate" = "999912"
              )
            )
          )
        )
      )
    )
  )
}

fixture_term_root <- function() {
  list(
    RESULT = list(status = "0", errorMsg = "Success."),
    METADATA_INF = list(
      CLASS_INF = list(
        CLASS_OBJ = list(
          CLASS = list(
            list(
              "@category" = "Population",
              "@name" = "Total population",
              "@code" = "0201010001",
              "@detail" = "Population definition.",
              "@linkURL" = "https://example.test/population"
            )
          )
        )
      )
    )
  )
}

fixture_event_root <- function() {
  list(
    RESULT = list(status = "0", errorMsg = "Success."),
    METADATA_INF = list(
      CLASS_INF = list(
        CLASS_OBJ = list(
          list(
            "@code" = "00003",
            "@name" = "Jinmu boom",
            "@level" = "3",
            "@fromTime" = "Dec.1954",
            "@toTime" = "Jun.1957",
            CLASS = list(
              list("@code" = "0201", "@name" = "Population")
            )
          )
        )
      )
    )
  )
}

fixture_survey_root <- function() {
  list(
    RESULT = list(status = "0", errorMsg = "Success."),
    METADATA_INF = list(
      CLASS_INF = list(
        CLASS_OBJ = list(
          CLASS = list(
            list(
              "@code" = "20020101",
              "@name" = "Population Census / Population Estimates",
              "@agency" = "Statistics Bureau of Japan",
              "@kind" = "Fundamental Statistics",
              "@summary" = "Population statistics.",
              "@linkUrl" = "https://example.test/census"
            )
          )
        )
      )
    )
  )
}

fixture_data_root <- function(value = "125000000") {
  class_object <- function(id, name, classes) {
    list("@id" = id, "@name" = name, CLASS = classes)
  }
  list(
    RESULT = list(status = "0", errorMsg = "Success."),
    STATISTICAL_DATA = list(
      RESULT_INF = list(TOTAL_NUMBER = "1"),
      TABLE_INF = list(
        STAT_NAME = list(
          list("@code" = "20020101", "$" = "Population Census")
        )
      ),
      CLASS_INF = list(
        CLASS_OBJ = list(
          class_object(
            "indicator", "Indicator",
            list(list(
              "@code" = "0201010000000010000",
              "@name" = "Total population (Both sexes)"
            ))
          ),
          class_object(
            "unit", "Unit",
            list(list("@code" = "090", "$" = "person"))
          ),
          class_object(
            "regionCode", "Region",
            list(list("@code" = "00000", "@name" = "Japan"))
          ),
          class_object(
            "time", "Time axis",
            list(list("@code" = "2024CY00", "@name" = "2024"))
          ),
          class_object(
            "cycle", "Cycle",
            list(list("@code" = "3", "$" = "Calendar Year"))
          ),
          class_object(
            "regionalRank", "Regional rank",
            list(list("@code" = "2", "$" = "Japan"))
          ),
          class_object(
            "isSeasonal", "Original or adjusted",
            list(list("@code" = "1", "$" = "Original Series"))
          ),
          class_object(
            "isProvisional", "Preliminary or fixed",
            list(list("@code" = "0", "$" = "Fixed"))
          )
        )
      ),
      DATA_INF = list(
        DATA_OBJ = list(
          list(
            VALUE = list(
              "@indicator" = "0201010000000010000",
              "@unit" = "090",
              "@stat" = "20020101",
              "@regionCode" = "00000",
              "@time" = "2024CY00",
              "@cycle" = "3",
              "@regionRank" = "2",
              "@isSeasonal" = "1",
              "@isProvisional" = "0",
              "$" = value
            ),
            CELL_ANNOTATIONS = list("$" = "Fixture note.")
          )
        )
      )
    )
  )
}

test_that("public metadata functions send exact API parameter names", {
  captured <- NULL
  local_mocked_bindings(
    dash_fetch_json = function(endpoint, params, ...) {
      captured <<- list(endpoint = endpoint, params = params)
      fixture_indicator_root()
    },
    .package = "dashboardapi"
  )
  out <- dashboard_search(
    "Total population",
    cycle = "year",
    regional_rank = "japan",
    seasonal = "original"
  )
  expect_equal(nrow(out), 1L)
  expect_identical(captured$endpoint, "getIndicatorInfo")
  expect_identical(captured$params$SearchIndicatorWord, "Total population")
  expect_identical(captured$params$Cycle, "3")
  expect_identical(captured$params$RegionalRank, "2")
  expect_identical(captured$params$IsSeasonalAdjustment, "1")
})

test_that("dashboard_data batches indicators and preserves aliases", {
  calls <- list()
  local_mocked_bindings(
    dash_fetch_json = function(endpoint, params, ...) {
      calls[[length(calls) + 1L]] <<- params
      fixture_data_root()
    },
    dash_pause = function(...) invisible(NULL),
    .package = "dashboardapi"
  )
  codes <- sprintf("%019d", 1:6)
  names(codes) <- paste0("series", 1:6)
  out <- dashboard_data(codes, wait = 0)
  expect_equal(length(calls), 2L)
  expect_equal(length(strsplit(calls[[1L]]$IndicatorCode, ",")[[1L]]), 5L)
  expect_identical(calls[[1L]]$MetaGetFlg, "Y")
  expect_s3_class(out, "tbl_df")
})

test_that("credit and common code tables are available offline", {
  credit <- dashboard_api_credit("jp")
  expect_identical(
    credit$source,
    "出典：統計ダッシュボード（https://dashboard.e-stat.go.jp/）"
  )
  expect_match(credit$processed, "<作成主体>", fixed = TRUE)
  expect_match(credit$credit, "統計ダッシュボード", fixed = TRUE)
  expect_match(
    dashboard_api_credit("en")$credit,
    "This service uses the API feature of Statistics Dashboard",
    fixed = TRUE
  )
  expect_equal(nrow(dashboard_codes("cycle")), 4L)
  expect_identical(
    dashboard_codes("seasonal", "jp")$name,
    c("原数値", "季節調整値")
  )
})

test_that("indicator metadata become indicator-element rows", {
  out <- dashboardapi:::dash_parse_indicators(
    fixture_indicator_root(), "EN"
  )
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)
  expect_identical(out$indicator_code, "0201010000000010000")
  expect_identical(
    out$indicator_element_code,
    "0201010000000010000030201"
  )
  expect_identical(out$cycle_name, "Calendar Year")
  expect_identical(out$annotation, "Sample annotation.")
})

test_that("region, term, event, and survey metadata are normalized", {
  region <- dashboardapi:::dash_parse_regions(fixture_region_root(), "EN")
  expect_identical(region$region_code, "13000")
  expect_identical(region$parent_region, "Japan")

  term <- dashboardapi:::dash_parse_simple_class(
    fixture_term_root(), "getTermInfo",
    c(name = "@name", code = "@code"), "EN"
  )
  expect_identical(term$code, "0201010001")

  event <- dashboardapi:::dash_parse_events(fixture_event_root(), "EN")
  expect_identical(event$event_code, "00003")
  expect_identical(event$category_code, "0201")

  survey <- dashboardapi:::dash_parse_simple_class(
    fixture_survey_root(), "getStatInfo",
    c(stat_code = "@code", name = "@name"), "EN"
  )
  expect_identical(survey$stat_code, "20020101")
})

test_that("data responses become normalized long data", {
  aliases <- c("0201010000000010000" = "population")
  out <- dashboardapi:::dash_parse_data(fixture_data_root(), aliases)
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)
  expect_identical(out$indicator, "population")
  expect_identical(out$region, "Japan")
  expect_equal(out$date, as.Date("2024-01-01"))
  expect_equal(out$value, 125000000)
  expect_false(out$provisional)
  expect_identical(out$annotation, "Fixture note.")
})

test_that("non-numeric API values are retained and normalized to NA", {
  out <- dashboardapi:::dash_parse_data(fixture_data_root("***"))
  expect_true(is.na(out$value))
  expect_identical(out$value_raw, "***")
})

test_that("wide data use indicator aliases", {
  aliases <- c("0201010000000010000" = "population")
  long <- dashboardapi:::dash_parse_data(fixture_data_root(), aliases)
  wide <- dashboardapi:::dash_widen_data(long)
  expect_true("population" %in% names(wide))
  expect_equal(wide$population, 125000000)
})

test_that("API periods become dates with correct fiscal semantics", {
  period_date <- dashboardapi:::dash_period_date
  expect_equal(period_date("20240100"), as.Date("2024-01-01"))
  expect_equal(period_date("20242Q00"), as.Date("2024-04-01"))
  expect_equal(period_date("2024CY00"), as.Date("2024-01-01"))
  expect_equal(period_date("2024FY00"), as.Date("2024-04-01"))
})

test_that("no-data responses return typed empty tibbles", {
  root <- fixture_indicator_root(status = "1")
  out <- dashboardapi:::dash_parse_indicators(root, "EN")
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
  expect_type(out$indicator_code, "character")
})

dash_abort <- function(message, class = "dashboardapi_error", ..., call = NULL) {
  fields <- list(...)
  condition <- c(list(message = message, call = call), fields)
  class(condition) <- unique(c(
    class, "dashboardapi_error", "error", "condition"
  ))
  stop(condition)
}

dash_warn_no_data <- function() {
  condition <- structure(
    list(
      message = paste0(
        "The Statistics Dashboard API completed successfully, ",
        "but no data matched the query."
      ),
      call = NULL
    ),
    class = c(
      "dashboard_no_data_warning", "dashboardapi_warning",
      "warning", "condition"
    )
  )
  warning(condition)
}

`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

dash_default <- function(option, default) {
  getOption(paste0("dashboardapi.", option), default)
}

dash_scalar_character <- function(x, arg, allow_null = FALSE) {
  if (allow_null && is.null(x)) {
    return(NULL)
  }
  if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x)) {
    dash_abort(
      sprintf("`%s` must be one non-empty character value.", arg),
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_scalar_number <- function(x, arg, min = -Inf, integer = FALSE) {
  if (
    !is.numeric(x) || length(x) != 1L || is.na(x) ||
      !is.finite(x) || x < min
  ) {
    dash_abort(
      sprintf(
        "`%s` must be one number greater than or equal to %s.",
        arg, min
      ),
      class = "dashboard_parameter_error"
    )
  }
  if (integer && x != floor(x)) {
    dash_abort(
      sprintf("`%s` must be a whole number.", arg),
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_flag <- function(x, arg) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    dash_abort(
      sprintf("`%s` must be TRUE or FALSE.", arg),
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_match_lang <- function(lang = dash_default("lang", "en")) {
  if (is.null(lang)) {
    lang <- "en"
  }
  lang <- tolower(dash_scalar_character(lang, "lang"))
  if (!lang %in% c("en", "jp")) {
    dash_abort(
      "`lang` must be either \"en\" or \"jp\".",
      class = "dashboard_parameter_error"
    )
  }
  toupper(lang)
}

dash_optional_character <- function(x, arg) {
  if (is.null(x)) {
    return(NULL)
  }
  dash_scalar_character(x, arg)
}

dash_character_vector <- function(
    x,
    arg,
    max = Inf,
    pattern = NULL,
    allow_null = TRUE) {
  if (is.null(x) && allow_null) {
    return(NULL)
  }
  if (
    !is.character(x) || length(x) < 1L || anyNA(x) ||
      any(!nzchar(x))
  ) {
    dash_abort(
      sprintf("`%s` must be a non-empty character vector.", arg),
      class = "dashboard_parameter_error"
    )
  }
  if (length(x) > max) {
    dash_abort(
      sprintf("`%s` may contain at most %s values.", arg, max),
      class = "dashboard_parameter_error"
    )
  }
  if (!is.null(pattern) && any(!grepl(pattern, unname(x)))) {
    dash_abort(
      sprintf("`%s` contains an invalid code.", arg),
      class = "dashboard_parameter_error"
    )
  }
  unname(x)
}

dash_indicator_codes <- function(x, max = Inf) {
  dash_character_vector(
    x, "indicator_code", max = max, pattern = "^[0-9]{19}$",
    allow_null = FALSE
  )
}

dash_region_codes <- function(x, max = Inf, allow_null = TRUE) {
  dash_character_vector(
    x, "region_code", max = max, pattern = "^([0-9]{5}|[A-Z]{3})$",
    allow_null = allow_null
  )
}

dash_join <- function(x) {
  if (is.null(x)) NULL else paste(x, collapse = ",")
}

dash_chunks <- function(x, size) {
  if (is.null(x)) {
    return(list(NULL))
  }
  split(x, ceiling(seq_along(x) / size))
}

dash_choice_code <- function(x, arg, choices, allow_null = TRUE) {
  if (is.null(x) && allow_null) {
    return(NULL)
  }
  if (is.numeric(x) && length(x) == 1L && !is.na(x)) {
    x <- as.character(x)
  }
  x <- tolower(dash_scalar_character(x, arg))
  aliases <- names(choices)
  if (x %in% aliases) {
    return(unname(choices[[x]]))
  }
  valid_codes <- unique(unname(choices))
  if (x %in% valid_codes) {
    return(x)
  }
  dash_abort(
    sprintf(
      "`%s` must be one of %s.",
      arg, paste(sprintf("\"%s\"", aliases), collapse = ", ")
    ),
    class = "dashboard_parameter_error"
  )
}

dash_cycle <- function(x, allow_null = TRUE) {
  dash_choice_code(
    x, "cycle",
    c(
      month = "1", quarter = "2", year = "3",
      calendar_year = "3", fiscal_year = "4"
    ),
    allow_null
  )
}

dash_regional_rank <- function(x, allow_null = TRUE) {
  dash_choice_code(
    x, "regional_rank",
    c(country = "1", japan = "2", prefecture = "3", municipality = "4"),
    allow_null
  )
}

dash_seasonal <- function(x, allow_null = TRUE) {
  dash_choice_code(
    x, "seasonal",
    c(original = "1", seasonally_adjusted = "2"),
    allow_null
  )
}

dash_social_level <- function(x, allow_null = TRUE) {
  dash_choice_code(
    x, "level",
    c(low = "1", medium = "2", high = "3"),
    allow_null
  )
}

dash_region_level <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  if (is.numeric(x)) {
    x <- as.character(x)
  }
  if (
    !is.character(x) || length(x) < 1L || anyNA(x) ||
      any(!grepl("^[1-9][0-9]*$", x))
  ) {
    dash_abort(
      "`region_level` must be a character or numeric vector of positive codes.",
      class = "dashboard_parameter_error"
    )
  }
  paste(x, collapse = ",")
}

dash_time_code <- function(x, arg, allow_latest = FALSE) {
  if (is.null(x)) {
    return(NULL)
  }
  if (is.numeric(x) && length(x) == 1L && !is.na(x)) {
    x <- format(x, scientific = FALSE, trim = TRUE)
  }
  x <- toupper(dash_scalar_character(x, arg))
  if (allow_latest && identical(x, "99999999")) {
    return(x)
  }
  valid <- grepl(
    "^([0-9]{4}(0[1-9]|1[0-2])00|[0-9]{4}[1-4]Q00|[0-9]{4}(CY|FY)00)$",
    x
  )
  if (!valid) {
    dash_abort(
      sprintf(
        paste0(
          "`%s` must use an API time code such as ",
          "\"20240100\", \"20241Q00\", \"2024CY00\", or \"2024FY00\"."
        ),
        arg
      ),
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_check_time_pair <- function(time_from, time_to) {
  if (is.null(time_from) || is.null(time_to)) {
    return(invisible(NULL))
  }
  type <- function(x) {
    if (grepl("Q00$", x)) "quarter"
    else if (grepl("CY00$", x)) "calendar_year"
    else if (grepl("FY00$", x)) "fiscal_year"
    else "month"
  }
  if (!identical(type(time_from), type(time_to))) {
    dash_abort(
      "`time_from` and `time_to` must use the same time-code format.",
      class = "dashboard_parameter_error"
    )
  }
  if (time_from > time_to) {
    dash_abort(
      "`time_to` must not be earlier than `time_from`.",
      class = "dashboard_parameter_error"
    )
  }
  invisible(NULL)
}

dash_date8 <- function(x, arg) {
  if (is.null(x)) {
    return(NULL)
  }
  x <- dash_scalar_character(x, arg)
  parsed <- suppressWarnings(as.Date(x, format = "%Y%m%d"))
  if (!grepl("^[0-9]{8}$", x) || is.na(parsed)) {
    dash_abort(
      sprintf("`%s` must be a valid date in YYYYMMDD format.", arg),
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_value_condition <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  x <- dash_scalar_character(x, "value_condition")
  if (!grepl("^[<>]-?(([0-9]+(\\.[0-9]*)?)|(\\.[0-9]+))$", x)) {
    dash_abort(
      "`value_condition` must begin with '<' or '>' followed by a number.",
      class = "dashboard_parameter_error"
    )
  }
  x
}

dash_records <- function(x) {
  if (is.null(x) || length(x) == 0L) {
    return(list())
  }
  if (!is.list(x)) {
    return(list(x))
  }
  if (is.null(names(x))) {
    return(x)
  }
  list(x)
}

dash_field <- function(x, name, default = "") {
  value <- x[[name]]
  if (is.null(value) || length(value) == 0L) {
    return(default)
  }
  value <- unlist(value, recursive = TRUE, use.names = FALSE)
  if (length(value) == 0L || is.na(value[[1L]])) {
    return(default)
  }
  as.character(value[[1L]])
}

dash_record_text <- function(records, field) {
  values <- vapply(
    dash_records(records),
    dash_field,
    character(1L),
    name = field,
    default = ""
  )
  values[nzchar(values)]
}

dash_bind_rows <- function(rows, empty) {
  rows <- Filter(Negate(is.null), rows)
  if (length(rows) == 0L) {
    return(empty)
  }
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  tibble::as_tibble(out)
}

dash_period_date <- function(x) {
  if (length(x) == 0L) {
    return(as.Date(character()))
  }
  vapply_date <- function(code) {
    if (is.na(code) || !nzchar(code)) {
      return(as.Date(NA))
    }
    if (grepl("^[0-9]{4}(0[1-9]|1[0-2])00$", code)) {
      return(as.Date(paste0(substr(code, 1L, 4L), "-", substr(code, 5L, 6L), "-01")))
    }
    if (grepl("^[0-9]{4}[1-4]Q00$", code)) {
      month <- (as.integer(substr(code, 5L, 5L)) - 1L) * 3L + 1L
      return(as.Date(sprintf("%s-%02d-01", substr(code, 1L, 4L), month)))
    }
    if (grepl("^[0-9]{4}CY00$", code)) {
      return(as.Date(paste0(substr(code, 1L, 4L), "-01-01")))
    }
    if (grepl("^[0-9]{4}FY00$", code)) {
      return(as.Date(paste0(substr(code, 1L, 4L), "-04-01")))
    }
    as.Date(NA)
  }
  as.Date(
    unname(vapply(x, vapply_date, as.Date(NA))),
    origin = "1970-01-01"
  )
}

dash_element_code <- function(indicator, cycle, rank, seasonal) {
  make_two <- function(x) {
    number <- suppressWarnings(as.integer(x))
    ifelse(is.na(number), "", sprintf("%02d", number))
  }
  ok <- nzchar(indicator) & nzchar(cycle) & nzchar(rank) & nzchar(seasonal)
  out <- rep("", length(indicator))
  out[ok] <- paste0(
    indicator[ok],
    make_two(cycle[ok]),
    make_two(rank[ok]),
    make_two(seasonal[ok])
  )
  out
}

dash_sleep <- function(seconds) {
  Sys.sleep(seconds)
}

dash_pause <- function(wait) {
  wait <- dash_scalar_number(wait, "wait", min = 0)
  dash_sleep(max(1, wait))
  invisible(NULL)
}

dash_drop_null <- function(x) {
  x[!vapply(x, is.null, logical(1L))]
}

#' Return Statistics Dashboard attribution and API credit
#'
#' The Statistics Dashboard asks users of its content to cite the source. When
#' publishing edited content, identify the editing entity as well. Publishers of
#' services using the API must also show a specified credit. The terms can
#' change, so verify the linked official page before publication.
#'
#' @param lang Credit language, `"en"` or `"jp"`.
#'
#' @return A named list containing `source` (the required source citation),
#'   `processed` (a template for edited content), `credit` (the required API
#'   service credit), and official URLs.
#' @export
#' @examples
#' dashboard_api_credit("en")
#' dashboard_api_credit("jp")$credit
dashboard_api_credit <- function(lang = c("en", "jp")) {
  lang <- match.arg(lang)
  if (identical(lang, "jp")) {
    return(list(
      source = paste0(
        "\u51FA\u5178\uFF1A\u7D71\u8A08\u30C0\u30C3\u30B7\u30E5\u30DC\u30FC\u30C9",
        "\uFF08https://dashboard.e-stat.go.jp/\uFF09"
      ),
      processed = paste0(
        "\u7D71\u8A08\u30C0\u30C3\u30B7\u30E5\u30DC\u30FC\u30C9",
        "\uFF08https://dashboard.e-stat.go.jp/\uFF09\u306E\u30C7\u30FC\u30BF\u3092",
        "\u52A0\u5DE5\u3057\u3066 <\u4F5C\u6210\u4E3B\u4F53> \u4F5C\u6210"
      ),
      credit = paste0(
        "\u3053\u306E\u30B5\u30FC\u30D3\u30B9\u306F\u3001",
        "\u7D71\u8A08\u30C0\u30C3\u30B7\u30E5\u30DC\u30FC\u30C9",
        "\u306EAPI\u6A5F\u80FD\u3092\u4F7F\u7528\u3057\u3066",
        "\u3044\u307E\u3059\u304C\u3001\u30B5\u30FC\u30D3\u30B9",
        "\u306E\u5185\u5BB9\u306F\u56FD\u306B\u3088\u3063\u3066",
        "\u4FDD\u8A3C\u3055\u308C\u305F\u3082\u306E\u3067\u306F",
        "\u3042\u308A\u307E\u305B\u3093\u3002"
      ),
      api_url = "https://dashboard.e-stat.go.jp/static/api",
      policy_url = "https://dashboard.e-stat.go.jp/static/sitePolicy",
      terms_url = "https://dashboard.e-stat.go.jp/static/terms"
    ))
  }
  list(
    source = "Source: Statistics Dashboard (https://dashboard.e-stat.go.jp/en/)",
    processed = paste0(
      "Created by <entity> based on the Statistics Dashboard ",
      "(https://dashboard.e-stat.go.jp/en/)"
    ),
    credit = paste0(
      "This service uses the API feature of Statistics Dashboard, ",
      "but the contents of this service are not guaranteed by the ",
      "Statistics Bureau of Japan."
    ),
    api_url = "https://dashboard.e-stat.go.jp/en/static/api",
    policy_url = "https://dashboard.e-stat.go.jp/en/static/sitePolicy",
    terms_url = "https://dashboard.e-stat.go.jp/en/static/terms"
  )
}

.dashboardapi_max_retry_after <- 60

dash_request <- function(endpoint, params, timeout) {
  url <- paste0(.dashboardapi_base_url, "/", endpoint)
  package_version <- tryCatch(
    as.character(utils::packageVersion("dashboardapi")),
    error = function(e) "development"
  )

  request <- httr2::request(url)
  request <- do.call(
    httr2::req_url_query,
    c(list(request), dash_drop_null(params))
  )
  request <- httr2::req_headers(request, `Accept-Encoding` = "gzip")
  request <- httr2::req_user_agent(
    request,
    paste0(
      "dashboardapi/", package_version,
      " (R; +https://github.com/kenjimyzk/dashboardapi)"
    )
  )
  request <- httr2::req_timeout(request, seconds = timeout)
  httr2::req_error(request, is_error = function(resp) FALSE)
}

dash_perform <- function(request) {
  httr2::req_perform(request)
}

dash_endpoint_root <- function(endpoint) {
  switch(
    endpoint,
    getIndicatorInfo = "GET_META_INDICATOR_INF",
    getRegionInfo = "GET_META_REGION_INF",
    getTermInfo = "GET_META_TERM_INFO",
    getSocialEventInfo = "GET_META_SOCIAL_INFO",
    getStatInfo = "GET_META_STAT_INFO",
    getData = "GET_STATS",
    dash_abort(
      sprintf("Unknown API endpoint: %s.", endpoint),
      class = "dashboard_internal_error"
    )
  )
}

dash_response_root <- function(payload, endpoint, http_status = NA_integer_) {
  error_class <- if (
    length(http_status) == 1L && !is.na(http_status) && http_status >= 400L
  ) {
    "dashboard_http_error"
  } else {
    "dashboard_parse_error"
  }
  if (!is.list(payload) || is.null(names(payload))) {
    dash_abort(
      "The Statistics Dashboard API response must be a JSON object.",
      class = error_class,
      endpoint = endpoint,
      http_status = http_status
    )
  }
  root_name <- dash_endpoint_root(endpoint)
  root <- payload[[root_name]]
  if (!is.list(root)) {
    dash_abort(
      sprintf("The API response does not contain `%s`.", root_name),
      class = error_class,
      endpoint = endpoint,
      http_status = http_status
    )
  }
  root
}

dash_response_status <- function(root, endpoint, http_status = NA_integer_) {
  status <- dash_field(root$RESULT, "status", default = NA_character_)
  numeric_status <- suppressWarnings(as.integer(status))
  if (is.na(numeric_status) || !status %in% c("0", "1", "101", "103", "104", "105", "106", "200", "299")) {
    dash_abort(
      "The API response contains an invalid result status.",
      class = if (
        length(http_status) == 1L && !is.na(http_status) &&
          http_status >= 400L
      ) "dashboard_http_error" else "dashboard_parse_error",
      endpoint = endpoint,
      http_status = http_status
    )
  }
  numeric_status
}

dash_check_response <- function(payload, endpoint, http_status = NA_integer_) {
  root <- dash_response_root(payload, endpoint, http_status)
  status <- dash_response_status(root, endpoint, http_status)
  if (status == 0L) {
    return(root)
  }
  if (status == 1L) {
    dash_warn_no_data()
    return(root)
  }
  message <- dash_field(root$RESULT, "errorMsg", "Unknown API error")
  dash_abort(
    sprintf("Statistics Dashboard API error %s: %s", status, message),
    class = "dashboard_api_response_error",
    status = status,
    api_message = message,
    endpoint = endpoint,
    http_status = http_status
  )
}

dash_retry_pause <- function(response, attempt, wait) {
  retry_after <- NA_real_
  if (!is.null(response)) {
    retry_after <- suppressWarnings(
      as.numeric(httr2::resp_header(response, "retry-after"))
    )
  }
  if (is.na(retry_after) || !is.finite(retry_after) || retry_after < 0) {
    retry_after <- 0
  }
  if (retry_after > .dashboardapi_max_retry_after) {
    dash_abort(
      sprintf(
        paste0(
          "The API requested a Retry-After delay of %s seconds, ",
          "exceeding the safety limit of %s seconds."
        ),
        retry_after, .dashboardapi_max_retry_after
      ),
      class = "dashboard_http_error",
      retry_after = retry_after
    )
  }
  exponential <- min(8, 2^(attempt - 1L))
  jitter <- if (wait > 0) stats::runif(1L, 0, min(0.25, wait / 4)) else 0
  dash_sleep(max(wait, retry_after, exponential) + jitter)
}

dash_fetch_json <- function(
    endpoint,
    params,
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3),
    wait = dash_default("wait", 1)) {
  timeout <- dash_scalar_number(timeout, "timeout", min = 0)
  retries <- dash_scalar_number(retries, "retries", min = 0, integer = TRUE)
  wait <- dash_scalar_number(wait, "wait", min = 0)
  transient_http <- c(408L, 429L, 500L, 502L, 503L, 504L)
  max_tries <- as.integer(retries) + 1L

  for (attempt in seq_len(max_tries)) {
    request <- dash_request(endpoint, params, timeout)
    result <- tryCatch(
      list(response = dash_perform(request), error = NULL),
      error = function(e) list(response = NULL, error = e)
    )
    if (!is.null(result$error)) {
      if (attempt < max_tries) {
        dash_retry_pause(NULL, attempt, wait)
        next
      }
      dash_abort(
        paste0("The Statistics Dashboard API request failed: ",
               conditionMessage(result$error)),
        class = "dashboard_http_error",
        parent = result$error,
        endpoint = endpoint
      )
    }

    response <- result$response
    http_status <- httr2::resp_status(response)
    if (http_status %in% transient_http && attempt < max_tries) {
      dash_retry_pause(response, attempt, wait)
      next
    }

    payload <- tryCatch(
      httr2::resp_body_json(
        response, check_type = FALSE, simplifyVector = FALSE
      ),
      error = function(e) {
        dash_abort(
          paste0("The API returned unreadable JSON: ", conditionMessage(e)),
          class = if (http_status >= 400L) {
            "dashboard_http_error"
          } else {
            "dashboard_parse_error"
          },
          parent = e,
          endpoint = endpoint,
          http_status = http_status
        )
      }
    )

    root <- dash_check_response(payload, endpoint, http_status)
    if (http_status >= 400L) {
      dash_abort(
        sprintf("The API returned HTTP status %s.", http_status),
        class = "dashboard_http_error",
        endpoint = endpoint,
        http_status = http_status
      )
    }
    return(root)
  }

  dash_abort(
    "The API retry loop ended unexpectedly.",
    class = "dashboard_http_error",
    endpoint = endpoint
  )
}

test_that("indicator and region codes are validated", {
  expect_error(
    dashboard_data("short"),
    class = "dashboard_parameter_error"
  )
  expect_error(
    dashboard_data("0201010000000010000", region_code = "Tokyo"),
    class = "dashboard_parameter_error"
  )
  expect_error(
    dashboard_indicators(indicator_code = rep("0201010000000010000", 51)),
    class = "dashboard_parameter_error"
  )
})

test_that("time and value filters are validated", {
  expect_error(
    dashboard_data(
      "0201010000000010000",
      time_from = "20240100",
      time_to = "2024CY00"
    ),
    class = "dashboard_parameter_error"
  )
  expect_error(
    dashboard_data(
      "0201010000000010000",
      value_condition = ">=100"
    ),
    class = "dashboard_parameter_error"
  )
  expect_identical(
    dashboardapi:::dash_time_code("20242Q00", "time"),
    "20242Q00"
  )
})

test_that("API errors retain a specific condition class", {
  payload <- list(
    GET_STATS = list(
      RESULT = list(
        status = "101",
        errorMsg = "IndicatorCode is mandatory input."
      )
    )
  )
  expect_error(
    dashboardapi:::dash_check_response(payload, "getData", 400L),
    class = "dashboard_api_response_error"
  )
})

test_that("unexpected response structures are parse errors", {
  expect_error(
    dashboardapi:::dash_response_root(list(other = list()), "getData"),
    class = "dashboard_parse_error"
  )
})

dash_class_objects <- function(statistical_data) {
  dash_records(statistical_data$CLASS_INF$CLASS_OBJ)
}

dash_class_lookup <- function(statistical_data, id) {
  objects <- dash_class_objects(statistical_data)
  positions <- vapply(
    objects, function(x) identical(dash_field(x, "@id"), id),
    logical(1L)
  )
  if (!any(positions)) {
    return(character())
  }
  classes <- dash_records(objects[[which(positions)[1L]]]$CLASS)
  codes <- vapply(
    classes, dash_field, character(1L), name = "@code", default = ""
  )
  labels <- vapply(classes, function(x) {
    dash_field(x, "@name", dash_field(x, "$"))
  }, character(1L))
  stats::setNames(labels, codes)
}

dash_stat_lookup <- function(statistical_data) {
  records <- dash_records(statistical_data$TABLE_INF$STAT_NAME)
  codes <- vapply(
    records, dash_field, character(1L), name = "@code", default = ""
  )
  labels <- vapply(
    records, dash_field, character(1L), name = "$", default = ""
  )
  stats::setNames(labels, codes)
}

dash_lookup <- function(map, code) {
  if (!nzchar(code) || length(map) == 0L || !code %in% names(map)) {
    return("")
  }
  unname(map[[code]])
}

dash_data_annotation <- function(object) {
  annotation <- object$CELL_ANNOTATIONS
  if (is.null(annotation)) {
    return("")
  }
  values <- c(
    dash_record_text(annotation, "$"),
    dash_record_text(annotation$CELL_ANNOTATION, "$")
  )
  paste(unique(values[nzchar(values)]), collapse = "; ")
}

dash_empty_data <- function() {
  tibble::tibble(
    indicator = character(),
    indicator_code = character(),
    indicator_element_code = character(),
    indicator_name = character(),
    unit_code = character(),
    unit = character(),
    stat_code = character(),
    stat_name = character(),
    region_code = character(),
    region = character(),
    time = character(),
    date = as.Date(character()),
    cycle = character(),
    cycle_name = character(),
    regional_rank = character(),
    regional_rank_name = character(),
    seasonal = character(),
    seasonal_name = character(),
    provisional = logical(),
    provisional_code = character(),
    provisional_name = character(),
    value = double(),
    value_raw = character(),
    annotation = character()
  )
}

dash_parse_data <- function(root, aliases = NULL) {
  empty <- dash_empty_data()
  if (dash_response_status(root, "getData") == 1L) {
    return(empty)
  }
  statistical_data <- root$STATISTICAL_DATA
  if (!is.list(statistical_data)) {
    dash_abort(
      "The API response does not contain statistical data.",
      class = "dashboard_parse_error"
    )
  }
  maps <- list(
    indicator = dash_class_lookup(statistical_data, "indicator"),
    unit = dash_class_lookup(statistical_data, "unit"),
    region = dash_class_lookup(statistical_data, "regionCode"),
    time = dash_class_lookup(statistical_data, "time"),
    cycle = dash_class_lookup(statistical_data, "cycle"),
    rank = dash_class_lookup(statistical_data, "regionalRank"),
    seasonal = dash_class_lookup(statistical_data, "isSeasonal"),
    provisional = dash_class_lookup(statistical_data, "isProvisional"),
    stat = dash_stat_lookup(statistical_data)
  )
  objects <- dash_records(statistical_data$DATA_INF$DATA_OBJ)
  rows <- list()
  index <- 0L
  for (object in objects) {
    values <- dash_records(object$VALUE)
    for (item in values) {
      indicator_code <- dash_field(item, "@indicator")
      unit_code <- dash_field(item, "@unit")
      stat_code <- dash_field(item, "@stat")
      region_code <- dash_field(item, "@regionCode")
      time <- dash_field(item, "@time")
      cycle <- dash_field(item, "@cycle")
      rank <- dash_field(item, "@regionRank")
      seasonal <- dash_field(item, "@isSeasonal")
      provisional <- dash_field(item, "@isProvisional")
      raw_value <- dash_field(item, "$")
      numeric_value <- suppressWarnings(as.numeric(raw_value))
      if (!nzchar(raw_value) || !is.finite(numeric_value)) {
        numeric_value <- NA_real_
      }
      alias <- indicator_code
      if (
        !is.null(aliases) && indicator_code %in% names(aliases) &&
          nzchar(aliases[[indicator_code]])
      ) {
        alias <- aliases[[indicator_code]]
      }
      index <- index + 1L
      rows[[index]] <- data.frame(
        indicator = alias,
        indicator_code = indicator_code,
        indicator_element_code = dash_element_code(
          indicator_code, cycle, rank, seasonal
        ),
        indicator_name = dash_lookup(maps$indicator, indicator_code),
        unit_code = unit_code,
        unit = dash_lookup(maps$unit, unit_code),
        stat_code = stat_code,
        stat_name = dash_lookup(maps$stat, stat_code),
        region_code = region_code,
        region = dash_lookup(maps$region, region_code),
        time = time,
        date = dash_period_date(time),
        cycle = cycle,
        cycle_name = dash_lookup(maps$cycle, cycle),
        regional_rank = rank,
        regional_rank_name = dash_lookup(maps$rank, rank),
        seasonal = seasonal,
        seasonal_name = dash_lookup(maps$seasonal, seasonal),
        provisional = identical(provisional, "1"),
        provisional_code = provisional,
        provisional_name = dash_lookup(maps$provisional, provisional),
        value = numeric_value,
        value_raw = raw_value,
        annotation = dash_data_annotation(object),
        stringsAsFactors = FALSE
      )
    }
  }
  dash_bind_rows(rows, empty)
}

dash_widen_data <- function(data) {
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      region_code = character(), region = character(),
      time = character(), date = as.Date(character())
    ))
  }
  labels <- unique(data$indicator)
  if (any(!nzchar(labels)) || anyDuplicated(labels)) {
    dash_abort(
      "Indicator aliases must be non-empty and unique for wide output.",
      class = "dashboard_parameter_error"
    )
  }
  id_columns <- c(
    "region_code", "region", "time", "date", "cycle", "cycle_name",
    "regional_rank", "regional_rank_name", "seasonal", "seasonal_name",
    "stat_code", "stat_name"
  )
  key_part <- function(x) {
    if (inherits(x, "Date")) as.character(x) else as.character(x)
  }
  make_key <- function(x) {
    do.call(
      paste,
      c(lapply(x[id_columns], key_part), list(sep = "\r"))
    )
  }
  observation_key <- paste(make_key(data), data$indicator, sep = "\r")
  if (anyDuplicated(observation_key)) {
    dash_abort(
      paste0(
        "The query returned more than one value for the same wide-output cell. ",
        "Add cycle, regional-rank, seasonal, or survey filters, or use long output."
      ),
      class = "dashboard_wide_error"
    )
  }
  id <- unique(data[id_columns])
  key <- make_key(id)
  for (label in labels) {
    subset <- data[data$indicator == label, , drop = FALSE]
    id[[label]] <- subset$value[match(key, make_key(subset))]
  }
  tibble::as_tibble(id)
}

#' Retrieve Statistics Dashboard observations
#'
#' `dashboard_data()` retrieves one or more indicators and returns one
#' indicator-region-time observation per row. The API permits five indicators
#' and 50 regions per request. Longer vectors are split automatically, with at
#' least one second between requests.
#'
#' Named `indicator_code` vectors create aliases in the `indicator` column and
#' in wide output, following the style of `WDI::WDI()`.
#'
#' @param indicator_code Character vector of 19-digit indicator codes. Names
#'   become aliases. Requests are automatically batched in groups of five.
#' @param region_code Optional vector of five-digit Japanese region codes or
#'   three-letter ISO country codes. Requests are automatically batched in
#'   groups of 50.
#' @param parent_region_code Optional five-digit parent-region code.
#' @param region_level Optional official region-level code.
#' @param time,time_from,time_to Optional API time codes. Use `"20240100"` for
#'   January 2024, `"20241Q00"` for 2024 Q1, `"2024CY00"` for calendar year
#'   2024, or `"2024FY00"` for fiscal year 2024.
#' @param cycle Data cycle: `"month"`, `"quarter"`, `"year"`, or
#'   `"fiscal_year"`; API codes 1--4 are also accepted.
#' @param regional_rank One of `"country"`, `"japan"`, `"prefecture"`, or
#'   `"municipality"`; API codes 1--4 are also accepted.
#' @param seasonal `"original"` or `"seasonally_adjusted"`; API codes 1 and 2
#'   are also accepted.
#' @param stat_name Optional partial match for a statistical survey name.
#' @param value_condition Optional strict numeric filter such as `">100"` or
#'   `"<0"`.
#' @param wide If `FALSE` (default), return normalized long data. If `TRUE`,
#'   put each indicator or alias in its own column.
#' @param lang Response language, `"en"` or `"jp"`.
#' @param wait Requested seconds between automatically batched requests.
#'   Values below one are treated as one when another request is required.
#' @param timeout Request timeout in seconds.
#' @param retries Number of retries for transient failures.
#'
#' @return A tibble. Long output retains raw codes and labels, an R `Date`,
#'   numeric `value`, original `value_raw`, and annotations.
#' @export
#' @examples
#' \dontrun{
#' population <- dashboard_data(
#'   indicator_code = c(population = "0201010000000010000"),
#'   region_code = "00000",
#'   time_from = "2020CY00",
#'   time_to = "2024CY00",
#'   cycle = "year",
#'   regional_rank = "japan",
#'   seasonal = "original"
#' )
#' }
dashboard_data <- function(
    indicator_code,
    region_code = NULL,
    parent_region_code = NULL,
    region_level = NULL,
    time = NULL,
    time_from = NULL,
    time_to = NULL,
    cycle = NULL,
    regional_rank = NULL,
    seasonal = NULL,
    stat_name = NULL,
    value_condition = NULL,
    wide = FALSE,
    lang = dash_default("lang", "en"),
    wait = dash_default("wait", 1),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  original_names <- names(indicator_code)
  codes <- dash_indicator_codes(indicator_code)
  if (anyDuplicated(codes)) {
    dash_abort(
      "`indicator_code` must not contain duplicates.",
      class = "dashboard_parameter_error"
    )
  }
  regions <- if (is.null(region_code)) {
    NULL
  } else {
    dash_region_codes(toupper(region_code))
  }
  if (!is.null(regions) && anyDuplicated(regions)) {
    dash_abort(
      "`region_code` must not contain duplicates.",
      class = "dashboard_parameter_error"
    )
  }
  if (!is.null(parent_region_code)) {
    parent_region_code <- dash_scalar_character(
      parent_region_code, "parent_region_code"
    )
    if (!grepl("^[0-9]{5}$", parent_region_code)) {
      dash_abort(
        "`parent_region_code` must be a five-digit code.",
        class = "dashboard_parameter_error"
      )
    }
  }
  time <- dash_time_code(time, "time", allow_latest = FALSE)
  time_from <- dash_time_code(time_from, "time_from")
  time_to <- dash_time_code(time_to, "time_to")
  dash_check_time_pair(time_from, time_to)
  wide <- dash_flag(wide, "wide")
  wait <- dash_scalar_number(wait, "wait", min = 0)
  timeout <- dash_scalar_number(timeout, "timeout", min = 0)
  retries <- dash_scalar_number(retries, "retries", min = 0, integer = TRUE)
  lang <- dash_match_lang(lang)

  aliases <- codes
  names(aliases) <- codes
  if (!is.null(original_names)) {
    use <- !is.na(original_names) & nzchar(original_names)
    aliases[codes[use]] <- original_names[use]
  }
  if (wide && anyDuplicated(unname(aliases))) {
    dash_abort(
      "Named `indicator_code` aliases must be unique for wide output.",
      class = "dashboard_parameter_error"
    )
  }

  indicator_chunks <- dash_chunks(codes, .dashboardapi_max_indicators)
  region_chunks <- dash_chunks(regions, .dashboardapi_max_regions)
  outputs <- list()
  request_number <- 0L
  for (indicator_chunk in indicator_chunks) {
    for (region_chunk in region_chunks) {
      request_number <- request_number + 1L
      if (request_number > 1L) {
        dash_pause(wait)
      }
      params <- list(
        Lang = lang,
        IndicatorCode = dash_join(indicator_chunk),
        RegionCode = dash_join(region_chunk),
        ParentRegionCode = parent_region_code,
        RegionLevel = dash_region_level(region_level),
        Time = time,
        TimeFrom = time_from,
        TimeTo = time_to,
        Cycle = dash_cycle(cycle),
        RegionalRank = dash_regional_rank(regional_rank),
        IsSeasonalAdjustment = dash_seasonal(seasonal),
        StatName = dash_optional_character(stat_name, "stat_name"),
        ValueCondition = dash_value_condition(value_condition),
        MetaGetFlg = "Y",
        SectionHeaderFlg = "1"
      )
      root <- dash_fetch_json(
        "getData", params, timeout = timeout, retries = retries, wait = wait
      )
      outputs[[request_number]] <- dash_parse_data(root, aliases)
    }
  }
  out <- if (length(outputs) == 0L) {
    dash_empty_data()
  } else {
    dash_bind_rows(lapply(outputs, as.data.frame), dash_empty_data())
  }
  if (nrow(out) > 0L) {
    key <- paste(
      out$indicator_code, out$region_code, out$time, out$cycle,
      out$regional_rank, out$seasonal, out$stat_code, sep = "\r"
    )
    duplicate <- duplicated(key)
    if (any(duplicate)) {
      same_value <- out$value[match(key[duplicate], key)] == out$value[duplicate]
      same_value[is.na(same_value)] <-
        is.na(out$value[match(key[duplicate], key)]) & is.na(out$value[duplicate])
      if (!all(same_value)) {
        dash_abort(
          "Batched API responses contained conflicting duplicate observations.",
          class = "dashboard_incomplete_download"
        )
      }
      out <- out[!duplicated(key), , drop = FALSE]
    }
  }
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- list(
    indicator_code = codes,
    region_code = regions,
    parent_region_code = parent_region_code,
    region_level = region_level,
    time = time,
    time_from = time_from,
    time_to = time_to,
    cycle = cycle,
    regional_rank = regional_rank,
    seasonal = seasonal,
    stat_name = stat_name,
    value_condition = value_condition,
    lang = tolower(lang)
  )
  if (wide) dash_widen_data(out) else tibble::as_tibble(out)
}

dash_common_params <- function(
    lang,
    modified_from = NULL,
    modified_to = NULL) {
  list(
    Lang = dash_match_lang(lang),
    modifiedFrom = dash_date8(modified_from, "modified_from"),
    modifiedTo = dash_date8(modified_to, "modified_to")
  )
}

dash_indicator_annotation <- function(annotations, cycle, rank, seasonal) {
  annotations <- dash_records(annotations)
  if (length(annotations) == 0L) {
    return("")
  }
  matches <- vapply(annotations, function(item) {
    identical(dash_field(item, "@cycle"), cycle) &&
      identical(dash_field(item, "@regionalRank"), rank) &&
      identical(dash_field(item, "@isSeasonal"), seasonal)
  }, logical(1L))
  text <- vapply(
    annotations[matches],
    dash_field,
    character(1L),
    name = "@annotation",
    default = ""
  )
  paste(unique(text[nzchar(text)]), collapse = "; ")
}

dash_parse_indicators <- function(root, lang) {
  empty <- tibble::tibble(
    indicator_code = character(),
    indicator_element_code = character(),
    name = character(),
    short_name = character(),
    cycle = character(),
    cycle_name = character(),
    regional_rank = character(),
    regional_rank_name = character(),
    seasonal = character(),
    seasonal_name = character(),
    unit = character(),
    from_time = character(),
    to_time = character(),
    stat_name = character(),
    term_codes = character(),
    term_names = character(),
    term_details = character(),
    annotation = character(),
    lang = character()
  )
  if (dash_response_status(root, "getIndicatorInfo") == 1L) {
    return(empty)
  }
  objects <- dash_records(
    root$METADATA_INF$CLASS_INF$CLASS_OBJ
  )
  rows <- list()
  index <- 0L
  for (object in objects) {
    indicator_code <- dash_field(object, "@code")
    details <- dash_records(object$details$detail)
    term_codes <- paste(
      dash_record_text(details, "@code"), collapse = ";"
    )
    term_names <- paste(
      dash_record_text(details, "@name"), collapse = "; "
    )
    term_details <- paste(
      dash_record_text(details, "$"), collapse = "; "
    )
    classes <- dash_records(object$CLASS)
    for (item in classes) {
      cycle <- dash_field(item$cycle, "@code")
      rank <- dash_field(item$RegionalRank, "@code")
      seasonal <- dash_field(item$IsSeasonal, "@code")
      index <- index + 1L
      rows[[index]] <- data.frame(
        indicator_code = indicator_code,
        indicator_element_code = dash_element_code(
          indicator_code, cycle, rank, seasonal
        ),
        name = dash_field(item, "@name", dash_field(object, "@name")),
        short_name = dash_field(item, "@sname"),
        cycle = cycle,
        cycle_name = dash_field(item$cycle, "@name"),
        regional_rank = rank,
        regional_rank_name = dash_field(item$RegionalRank, "@name"),
        seasonal = seasonal,
        seasonal_name = dash_field(item$IsSeasonal, "@name"),
        unit = dash_field(item, "@unit"),
        from_time = dash_field(item, "@fromDate"),
        to_time = dash_field(item, "@toDate"),
        stat_name = dash_field(item, "@statName"),
        term_codes = term_codes,
        term_names = term_names,
        term_details = term_details,
        annotation = dash_indicator_annotation(
          object$annotations, cycle, rank, seasonal
        ),
        lang = tolower(lang),
        stringsAsFactors = FALSE
      )
    }
  }
  dash_bind_rows(rows, empty)
}

#' Search and retrieve indicator metadata
#'
#' This is the discovery endpoint for Statistics Dashboard series. The result
#' has one row per indicator element: an indicator split by data cycle,
#' regional rank, and original/seasonally adjusted status.
#'
#' @param query Optional partial match for the indicator name.
#' @param indicator_code Optional vector of 19-digit indicator codes. Up to 50
#'   codes can be sent in one metadata request.
#' @param category Optional Statistics Dashboard category code.
#' @param time,time_from,time_to Optional eight-character API time codes, such
#'   as `"20240100"`, `"20241Q00"`, `"2024CY00"`, or `"2024FY00"`.
#' @param cycle Data cycle: `"month"`, `"quarter"`, `"year"`, or
#'   `"fiscal_year"`; API codes 1--4 are also accepted.
#' @param regional_rank One of `"country"`, `"japan"`, `"prefecture"`, or
#'   `"municipality"`; API codes 1--4 are also accepted.
#' @param seasonal `"original"` or `"seasonally_adjusted"`; API codes 1 and 2
#'   are also accepted.
#' @param stat_code,stat_name Optional statistical survey code or partial name.
#' @param modified_from,modified_to Optional metadata update dates in
#'   `YYYYMMDD` format.
#' @param lang Response language, `"en"` or `"jp"`.
#' @param timeout Request timeout in seconds.
#' @param retries Number of retries for transient failures.
#'
#' @return A tibble with one row per indicator element.
#' @export
#' @examples
#' \dontrun{
#' dashboard_indicators(query = "Total population", lang = "en")
#' dashboard_indicators(
#'   category = "0201", cycle = "year", regional_rank = "prefecture"
#' )
#' }
dashboard_indicators <- function(
    query = NULL,
    indicator_code = NULL,
    category = NULL,
    time = NULL,
    time_from = NULL,
    time_to = NULL,
    cycle = NULL,
    regional_rank = NULL,
    seasonal = NULL,
    stat_code = NULL,
    stat_name = NULL,
    modified_from = NULL,
    modified_to = NULL,
    lang = dash_default("lang", "en"),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  codes <- if (is.null(indicator_code)) {
    NULL
  } else {
    dash_character_vector(
      indicator_code, "indicator_code", max = 50L,
      pattern = "^[0-9]{19}$", allow_null = TRUE
    )
  }
  time <- dash_time_code(time, "time", allow_latest = FALSE)
  time_from <- dash_time_code(time_from, "time_from")
  time_to <- dash_time_code(time_to, "time_to")
  dash_check_time_pair(time_from, time_to)
  lang <- dash_match_lang(lang)
  params <- c(
    list(
      Lang = lang,
      IndicatorCode = dash_join(codes),
      Category = dash_optional_character(category, "category"),
      Time = time,
      TimeFrom = time_from,
      TimeTo = time_to,
      Cycle = dash_cycle(cycle),
      RegionalRank = dash_regional_rank(regional_rank),
      IsSeasonalAdjustment = dash_seasonal(seasonal),
      StatCode = dash_optional_character(stat_code, "stat_code"),
      StatName = dash_optional_character(stat_name, "stat_name"),
      SearchIndicatorWord = dash_optional_character(query, "query"),
      modifiedFrom = dash_date8(modified_from, "modified_from"),
      modifiedTo = dash_date8(modified_to, "modified_to")
    )
  )
  root <- dash_fetch_json(
    "getIndicatorInfo", params, timeout = timeout, retries = retries
  )
  out <- dash_parse_indicators(root, lang)
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- dash_drop_null(params)
  out
}

#' Search Statistics Dashboard indicators
#'
#' A concise alias for [dashboard_indicators()] with a required search phrase.
#'
#' @param query Partial indicator-name match.
#' @inheritParams dashboard_indicators
#' @param ... Additional filters passed to [dashboard_indicators()].
#' @return A tibble with one row per matching indicator element.
#' @export
#' @examples
#' \dontrun{
#' dashboard_search("unemployment rate")
#' dashboard_search(intToUtf8(c(23436, 20840, 22833, 26989, 29575)), lang = "jp")
#' }
dashboard_search <- function(
    query,
    lang = dash_default("lang", "en"),
    ...) {
  query <- dash_scalar_character(query, "query")
  dashboard_indicators(query = query, lang = lang, ...)
}

dash_parse_regions <- function(root, lang) {
  empty <- tibble::tibble(
    parent_region_code = character(),
    parent_region = character(),
    parent_hiragana = character(),
    region_code = character(),
    region = character(),
    level = character(),
    hiragana = character(),
    from_time = character(),
    to_time = character(),
    lang = character()
  )
  if (dash_response_status(root, "getRegionInfo") == 1L) {
    return(empty)
  }
  objects <- dash_records(root$METADATA_INF$CLASS_INF$CLASS_OBJ)
  rows <- list()
  index <- 0L
  for (object in objects) {
    children <- dash_records(object$CLASS)
    for (child in children) {
      index <- index + 1L
      rows[[index]] <- data.frame(
        parent_region_code = dash_field(object, "@parentRegionCode"),
        parent_region = dash_field(object, "@name"),
        parent_hiragana = dash_field(object, "@hiragana"),
        region_code = dash_field(child, "@regionCode"),
        region = dash_field(child, "@name", dash_field(child, "$")),
        level = dash_field(child, "@level"),
        hiragana = dash_field(child, "@hiragana"),
        from_time = dash_field(child, "@fromDate"),
        to_time = dash_field(child, "@toDate"),
        lang = tolower(lang),
        stringsAsFactors = FALSE
      )
    }
  }
  dash_bind_rows(rows, empty)
}

#' Search and retrieve region metadata
#'
#' @param query Optional partial match for a region name.
#' @param region_code Optional vector of up to 50 five-digit Japanese region
#'   codes or three-letter ISO country codes.
#' @param parent_region_code Optional five-digit parent-region code. For
#'   example, `"00000"` returns Japan's prefectures.
#' @param region_level Optional vector of official region-level codes.
#' @inheritParams dashboard_indicators
#' @return A tibble of regions and their parent regions.
#' @export
#' @examples
#' \dontrun{
#' dashboard_regions(parent_region_code = "00000")
#' dashboard_regions(query = "Tokyo", lang = "en")
#' }
dashboard_regions <- function(
    query = NULL,
    region_code = NULL,
    parent_region_code = NULL,
    time = NULL,
    time_from = NULL,
    time_to = NULL,
    region_level = NULL,
    modified_from = NULL,
    modified_to = NULL,
    lang = dash_default("lang", "en"),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  codes <- if (is.null(region_code)) {
    NULL
  } else {
    dash_region_codes(region_code, max = 50L)
  }
  if (!is.null(parent_region_code)) {
    parent_region_code <- dash_scalar_character(
      parent_region_code, "parent_region_code"
    )
    if (!grepl("^[0-9]{5}$", parent_region_code)) {
      dash_abort(
        "`parent_region_code` must be a five-digit code.",
        class = "dashboard_parameter_error"
      )
    }
  }
  time <- dash_time_code(time, "time")
  time_from <- dash_time_code(time_from, "time_from")
  time_to <- dash_time_code(time_to, "time_to")
  dash_check_time_pair(time_from, time_to)
  lang <- dash_match_lang(lang)
  params <- list(
    Lang = lang,
    RegionCode = dash_join(codes),
    ParentRegionCode = parent_region_code,
    Time = time,
    TimeFrom = time_from,
    TimeTo = time_to,
    RegionLevel = dash_region_level(region_level),
    SearchRegionWord = dash_optional_character(query, "query"),
    modifiedFrom = dash_date8(modified_from, "modified_from"),
    modifiedTo = dash_date8(modified_to, "modified_to")
  )
  root <- dash_fetch_json(
    "getRegionInfo", params, timeout = timeout, retries = retries
  )
  out <- dash_parse_regions(root, lang)
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- dash_drop_null(params)
  out
}

dash_parse_simple_class <- function(root, endpoint, fields, lang) {
  empty_values <- lapply(fields, function(x) character())
  names(empty_values) <- names(fields)
  empty_values$lang <- character()
  empty <- tibble::as_tibble(empty_values)
  if (dash_response_status(root, endpoint) == 1L) {
    return(empty)
  }
  containers <- dash_records(root$METADATA_INF$CLASS_INF$CLASS_OBJ)
  records <- unlist(
    lapply(containers, function(x) dash_records(x$CLASS)),
    recursive = FALSE
  )
  rows <- lapply(records, function(item) {
    values <- lapply(fields, function(field) dash_field(item, field))
    values$lang <- tolower(lang)
    as.data.frame(values, stringsAsFactors = FALSE)
  })
  dash_bind_rows(rows, empty)
}

#' Search and retrieve statistical term metadata
#'
#' @param query Optional partial match for a term name.
#' @param category,indicator_code,stat_code Optional metadata filters.
#' @inheritParams dashboard_indicators
#' @return A tibble of statistical terms and definitions.
#' @export
#' @examples
#' \dontrun{
#' dashboard_terms(category = "0201", lang = "en")
#' dashboard_terms(query = intToUtf8(c(20154, 21475)), lang = "jp")
#' }
dashboard_terms <- function(
    query = NULL,
    category = NULL,
    indicator_code = NULL,
    stat_code = NULL,
    modified_from = NULL,
    modified_to = NULL,
    lang = dash_default("lang", "en"),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  if (!is.null(indicator_code)) {
    indicator_code <- dash_character_vector(
      indicator_code, "indicator_code", max = 1L,
      pattern = "^[0-9]{19}$"
    )
  }
  lang <- dash_match_lang(lang)
  params <- list(
    Lang = lang,
    Category = dash_optional_character(category, "category"),
    IndicatorCode = dash_join(indicator_code),
    StatCode = dash_optional_character(stat_code, "stat_code"),
    SearchTermWord = dash_optional_character(query, "query"),
    modifiedFrom = dash_date8(modified_from, "modified_from"),
    modifiedTo = dash_date8(modified_to, "modified_to")
  )
  root <- dash_fetch_json(
    "getTermInfo", params, timeout = timeout, retries = retries
  )
  out <- dash_parse_simple_class(
    root, "getTermInfo",
    c(
      category = "@category", name = "@name", code = "@code",
      detail = "@detail", link_url = "@linkURL"
    ),
    lang
  )
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- dash_drop_null(params)
  out
}

dash_parse_events <- function(root, lang) {
  empty <- tibble::tibble(
    event_code = character(),
    event = character(),
    level = character(),
    from_time = character(),
    to_time = character(),
    category_code = character(),
    category = character(),
    lang = character()
  )
  if (dash_response_status(root, "getSocialEventInfo") == 1L) {
    return(empty)
  }
  events <- dash_records(root$METADATA_INF$CLASS_INF$CLASS_OBJ)
  rows <- list()
  index <- 0L
  for (event in events) {
    categories <- dash_records(event$CLASS)
    if (length(categories) == 0L) {
      categories <- list(list())
    }
    for (category in categories) {
      index <- index + 1L
      rows[[index]] <- data.frame(
        event_code = dash_field(event, "@code"),
        event = dash_field(event, "@name"),
        level = dash_field(event, "@level"),
        from_time = dash_field(event, "@fromTime"),
        to_time = dash_field(event, "@toTime"),
        category_code = dash_field(category, "@code"),
        category = dash_field(category, "@name", dash_field(category, "$")),
        lang = tolower(lang),
        stringsAsFactors = FALSE
      )
    }
  }
  dash_bind_rows(rows, empty)
}

#' Search and retrieve social-event metadata
#'
#' @param level Event importance: `"low"`, `"medium"`, or `"high"`; API
#'   codes 1--3 are also accepted.
#' @param category Optional Statistics Dashboard category code.
#' @inheritParams dashboard_indicators
#' @return A tibble with one row per event-category association.
#' @export
#' @examples
#' \dontrun{
#' dashboard_events(category = "0201", level = "high")
#' }
dashboard_events <- function(
    time = NULL,
    time_from = NULL,
    time_to = NULL,
    level = NULL,
    category = NULL,
    modified_from = NULL,
    modified_to = NULL,
    lang = dash_default("lang", "en"),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  time <- dash_time_code(time, "time")
  time_from <- dash_time_code(time_from, "time_from")
  time_to <- dash_time_code(time_to, "time_to")
  dash_check_time_pair(time_from, time_to)
  lang <- dash_match_lang(lang)
  params <- list(
    Lang = lang,
    Time = time,
    TimeFrom = time_from,
    TimeTo = time_to,
    SocialEventLevel = dash_social_level(level),
    Category = dash_optional_character(category, "category"),
    modifiedFrom = dash_date8(modified_from, "modified_from"),
    modifiedTo = dash_date8(modified_to, "modified_to")
  )
  root <- dash_fetch_json(
    "getSocialEventInfo", params, timeout = timeout, retries = retries
  )
  out <- dash_parse_events(root, lang)
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- dash_drop_null(params)
  out
}

#' Search and retrieve statistical-survey metadata
#'
#' @param query Optional partial match for a survey name.
#' @param indicator_code Optional vector of up to 50 indicator codes.
#' @param stat_code Optional statistical survey code.
#' @inheritParams dashboard_indicators
#' @return A tibble of statistical surveys.
#' @export
#' @examples
#' \dontrun{
#' dashboard_surveys(query = "Population Census", lang = "en")
#' }
dashboard_surveys <- function(
    query = NULL,
    indicator_code = NULL,
    stat_code = NULL,
    modified_from = NULL,
    modified_to = NULL,
    lang = dash_default("lang", "en"),
    timeout = dash_default("timeout", 30),
    retries = dash_default("retries", 3)) {
  codes <- if (is.null(indicator_code)) {
    NULL
  } else {
    dash_character_vector(
      indicator_code, "indicator_code", max = 50L,
      pattern = "^[0-9]{19}$"
    )
  }
  lang <- dash_match_lang(lang)
  params <- list(
    Lang = lang,
    IndicatorCode = dash_join(codes),
    StatCode = dash_optional_character(stat_code, "stat_code"),
    SearchSurveyWord = dash_optional_character(query, "query"),
    modifiedFrom = dash_date8(modified_from, "modified_from"),
    modifiedTo = dash_date8(modified_to, "modified_to")
  )
  root <- dash_fetch_json(
    "getStatInfo", params, timeout = timeout, retries = retries
  )
  out <- dash_parse_simple_class(
    root, "getStatInfo",
    c(
      stat_code = "@code", name = "@name", agency = "@agency",
      kind = "@kind", summary = "@summary", link_url = "@linkUrl"
    ),
    lang
  )
  attr(out, "retrieved_at") <- Sys.time()
  attr(out, "query") <- dash_drop_null(params)
  out
}

test_that("small live API requests match the documented schema", {
  skip_on_cran()
  skip_if_not(
    identical(tolower(Sys.getenv("DASHBOARDAPI_RUN_LIVE_TESTS")), "true")
  )

  indicators <- dashboard_search(
    "Total population (Both sexes)", lang = "en"
  )
  expect_s3_class(indicators, "tbl_df")
  expect_true("0201010000000010000" %in% indicators$indicator_code)

  Sys.sleep(1)
  regions <- dashboard_regions(
    parent_region_code = "00000", lang = "en"
  )
  expect_true(all(c("01000", "13000", "47000") %in% regions$region_code))

  Sys.sleep(1)
  terms <- dashboard_terms(category = "0201", lang = "en")
  expect_true("0201010001" %in% terms$code)

  Sys.sleep(1)
  events <- dashboard_events(category = "0201", lang = "en")
  expect_gt(nrow(events), 10L)
  expect_true(all(nzchar(events$event_code)))

  Sys.sleep(1)
  surveys <- dashboard_surveys(
    stat_code = "10070102", lang = "en"
  )
  expect_identical(
    surveys$stat_code,
    "10070102"
  )

  Sys.sleep(1)
  data <- dashboard_data(
    c(population = "0201010000000010000"),
    region_code = "00000",
    time_from = "2020CY00",
    time_to = "2021CY00",
    cycle = "year",
    regional_rank = "japan",
    seasonal = "original",
    lang = "en"
  )
  expect_equal(nrow(data), 2L)
  expect_identical(unique(data$indicator), "population")
  expect_true(all(is.finite(data$value)))
})

---
title: "Getting started with dashboardapi"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with dashboardapi}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)
```

`dashboardapi` implements a three-step workflow for Japan's Statistics
Dashboard: discover an indicator, identify regions, and retrieve observations.
No API key is required.

## Discover an indicator

```{r}
library(dashboardapi)

meta <- dashboard_search(
  "Total population (Both sexes)",
  lang = "en"
)

meta[, c(
  "indicator_code", "name", "cycle_name", "regional_rank_name",
  "seasonal_name", "unit", "from_time", "to_time"
)]
```

The 19-digit `indicator_code` identifies a statistical concept. An indicator
can have several *indicator elements*, distinguished by cycle, regional rank,
and original/seasonally adjusted status. The package retains both codes:
`indicator_code` and the derived 25-digit `indicator_element_code`.

## Find region codes

```{r}
prefectures <- dashboard_regions(
  parent_region_code = "00000",
  lang = "en"
)

prefectures[, c("region_code", "region", "level")]
```

For Japanese geography, parent code `"00000"` returns prefectures. A
prefecture code can then be used as `parent_region_code` to discover its
municipalities. Country-level series use three-letter ISO codes.

## Retrieve observations

```{r}
population <- dashboard_data(
  indicator_code = c(population = "0201010000000010000"),
  region_code = "00000",
  time_from = "2020CY00",
  time_to = "2024CY00",
  cycle = "year",
  regional_rank = "japan",
  seasonal = "original",
  lang = "en"
)

population[, c("indicator", "region", "time", "date", "unit", "value")]
```

Names on `indicator_code` create aliases. Long output retains the API's raw
dimension codes and labels, the original value text, parsed numeric values,
provisional status, and cell annotations.

## Time codes

The API uses eight-character period codes:

| Meaning | API code | `date` |
|---|---|---|
| January 2024 | `20240100` | 2024-01-01 |
| 2024 Q2 | `20242Q00` | 2024-04-01 |
| Calendar year 2024 | `2024CY00` | 2024-01-01 |
| Fiscal year 2024 | `2024FY00` | 2024-04-01 |

`time` is authoritative; `date` is a convenient first-day representation.

## Wide output

```{r}
wide <- dashboard_data(
  indicator_code = c(
    population = "0201010000000010000",
    japanese_population = "0201020000000010000"
  ),
  region_code = "00000",
  time_from = "2020CY00",
  time_to = "2024CY00",
  cycle = "year",
  regional_rank = "japan",
  seasonal = "original",
  wide = TRUE
)
```

If a query would produce more than one value for a wide-output cell, the
package asks you to add cycle, regional-rank, seasonal, or survey filters.

## Auxiliary metadata

```{r}
dashboard_terms(category = "0201", lang = "en")
dashboard_events(category = "0201", level = "high", lang = "en")
dashboard_surveys(query = "Population Census", lang = "en")
dashboard_codes("cycle")
```

## Responsible API use and attribution

The official page asks users not to produce a large access volume in a short
period. `dashboard_data()` batches vectors longer than the documented
per-request limits and waits at least one second between calls.

When publishing content retrieved from the Statistics Dashboard, cite the
source. If content or data are edited, identify both the editing and its
entity. A published service using the API must also display the specified API
credit. `dashboard_api_credit()` returns all three strings; replace `<entity>`
before publishing edited content:

```{r}
attribution <- dashboard_api_credit("en")
attribution$source
attribution$processed
attribution$credit
```

Individual rights statements, including those for third-party content, take
precedence over the general public-data terms. Review the current official
terms before publication.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_terms}
\alias{dashboard_terms}
\title{Search and retrieve statistical term metadata}
\usage{
dashboard_terms(
  query = NULL,
  category = NULL,
  indicator_code = NULL,
  stat_code = NULL,
  modified_from = NULL,
  modified_to = NULL,
  lang = dash_default("lang", "en"),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{query}{Optional partial match for a term name.}

\item{category, indicator_code, stat_code}{Optional metadata filters.}

\item{modified_from, modified_to}{Optional metadata update dates in
\code{YYYYMMDD} format.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble of statistical terms and definitions.
}
\description{
Search and retrieve statistical term metadata
}
\examples{
\dontrun{
dashboard_terms(category = "0201", lang = "en")
dashboard_terms(query = intToUtf8(c(20154, 21475)), lang = "jp")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_surveys}
\alias{dashboard_surveys}
\title{Search and retrieve statistical-survey metadata}
\usage{
dashboard_surveys(
  query = NULL,
  indicator_code = NULL,
  stat_code = NULL,
  modified_from = NULL,
  modified_to = NULL,
  lang = dash_default("lang", "en"),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{query}{Optional partial match for a survey name.}

\item{indicator_code}{Optional vector of up to 50 indicator codes.}

\item{stat_code}{Optional statistical survey code.}

\item{modified_from, modified_to}{Optional metadata update dates in
\code{YYYYMMDD} format.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble of statistical surveys.
}
\description{
Search and retrieve statistical-survey metadata
}
\examples{
\dontrun{
dashboard_surveys(query = "Population Census", lang = "en")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_indicators}
\alias{dashboard_indicators}
\title{Search and retrieve indicator metadata}
\usage{
dashboard_indicators(
  query = NULL,
  indicator_code = NULL,
  category = NULL,
  time = NULL,
  time_from = NULL,
  time_to = NULL,
  cycle = NULL,
  regional_rank = NULL,
  seasonal = NULL,
  stat_code = NULL,
  stat_name = NULL,
  modified_from = NULL,
  modified_to = NULL,
  lang = dash_default("lang", "en"),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{query}{Optional partial match for the indicator name.}

\item{indicator_code}{Optional vector of 19-digit indicator codes. Up to 50
codes can be sent in one metadata request.}

\item{category}{Optional Statistics Dashboard category code.}

\item{time, time_from, time_to}{Optional eight-character API time codes, such
as \code{"20240100"}, \code{"20241Q00"}, \code{"2024CY00"}, or \code{"2024FY00"}.}

\item{cycle}{Data cycle: \code{"month"}, \code{"quarter"}, \code{"year"}, or
\code{"fiscal_year"}; API codes 1--4 are also accepted.}

\item{regional_rank}{One of \code{"country"}, \code{"japan"}, \code{"prefecture"}, or
\code{"municipality"}; API codes 1--4 are also accepted.}

\item{seasonal}{\code{"original"} or \code{"seasonally_adjusted"}; API codes 1 and 2
are also accepted.}

\item{stat_code, stat_name}{Optional statistical survey code or partial name.}

\item{modified_from, modified_to}{Optional metadata update dates in
\code{YYYYMMDD} format.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble with one row per indicator element.
}
\description{
This is the discovery endpoint for Statistics Dashboard series. The result
has one row per indicator element: an indicator split by data cycle,
regional rank, and original/seasonally adjusted status.
}
\examples{
\dontrun{
dashboard_indicators(query = "Total population", lang = "en")
dashboard_indicators(
  category = "0201", cycle = "year", regional_rank = "prefecture"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_search}
\alias{dashboard_search}
\title{Search Statistics Dashboard indicators}
\usage{
dashboard_search(query, lang = dash_default("lang", "en"), ...)
}
\arguments{
\item{query}{Partial indicator-name match.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{...}{Additional filters passed to \code{\link[=dashboard_indicators]{dashboard_indicators()}}.}
}
\value{
A tibble with one row per matching indicator element.
}
\description{
A concise alias for \code{\link[=dashboard_indicators]{dashboard_indicators()}} with a required search phrase.
}
\examples{
\dontrun{
dashboard_search("unemployment rate")
dashboard_search(intToUtf8(c(23436, 20840, 22833, 26989, 29575)), lang = "jp")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_events}
\alias{dashboard_events}
\title{Search and retrieve social-event metadata}
\usage{
dashboard_events(
  time = NULL,
  time_from = NULL,
  time_to = NULL,
  level = NULL,
  category = NULL,
  modified_from = NULL,
  modified_to = NULL,
  lang = dash_default("lang", "en"),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{time, time_from, time_to}{Optional eight-character API time codes, such
as \code{"20240100"}, \code{"20241Q00"}, \code{"2024CY00"}, or \code{"2024FY00"}.}

\item{level}{Event importance: \code{"low"}, \code{"medium"}, or \code{"high"}; API
codes 1--3 are also accepted.}

\item{category}{Optional Statistics Dashboard category code.}

\item{modified_from, modified_to}{Optional metadata update dates in
\code{YYYYMMDD} format.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble with one row per event-category association.
}
\description{
Search and retrieve social-event metadata
}
\examples{
\dontrun{
dashboard_events(category = "0201", level = "high")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata.R
\name{dashboard_regions}
\alias{dashboard_regions}
\title{Search and retrieve region metadata}
\usage{
dashboard_regions(
  query = NULL,
  region_code = NULL,
  parent_region_code = NULL,
  time = NULL,
  time_from = NULL,
  time_to = NULL,
  region_level = NULL,
  modified_from = NULL,
  modified_to = NULL,
  lang = dash_default("lang", "en"),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{query}{Optional partial match for a region name.}

\item{region_code}{Optional vector of up to 50 five-digit Japanese region
codes or three-letter ISO country codes.}

\item{parent_region_code}{Optional five-digit parent-region code. For
example, \code{"00000"} returns Japan's prefectures.}

\item{time, time_from, time_to}{Optional eight-character API time codes, such
as \code{"20240100"}, \code{"20241Q00"}, \code{"2024CY00"}, or \code{"2024FY00"}.}

\item{region_level}{Optional vector of official region-level codes.}

\item{modified_from, modified_to}{Optional metadata update dates in
\code{YYYYMMDD} format.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble of regions and their parent regions.
}
\description{
Search and retrieve region metadata
}
\examples{
\dontrun{
dashboard_regions(parent_region_code = "00000")
dashboard_regions(query = "Tokyo", lang = "en")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codes.R
\name{dashboard_codes}
\alias{dashboard_codes}
\title{Return commonly used Statistics Dashboard API codes}
\usage{
dashboard_codes(
  type = c("cycle", "regional_rank", "seasonal", "provisional", "social_event_level"),
  lang = c("en", "jp")
)
}
\arguments{
\item{type}{One of \code{"cycle"}, \code{"regional_rank"}, \code{"seasonal"},
\code{"provisional"}, or \code{"social_event_level"}.}

\item{lang}{Label language, \code{"en"} or \code{"jp"}.}
}
\value{
A tibble containing \code{code}, \code{name}, and \code{type}.
}
\description{
The API's full category, survey, and region code systems are maintained by
the Statistics Bureau. This helper returns the small enumerations used
directly by package arguments.
}
\examples{
dashboard_codes("cycle")
dashboard_codes("regional_rank", lang = "jp")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{dashboard_data}
\alias{dashboard_data}
\title{Retrieve Statistics Dashboard observations}
\usage{
dashboard_data(
  indicator_code,
  region_code = NULL,
  parent_region_code = NULL,
  region_level = NULL,
  time = NULL,
  time_from = NULL,
  time_to = NULL,
  cycle = NULL,
  regional_rank = NULL,
  seasonal = NULL,
  stat_name = NULL,
  value_condition = NULL,
  wide = FALSE,
  lang = dash_default("lang", "en"),
  wait = dash_default("wait", 1),
  timeout = dash_default("timeout", 30),
  retries = dash_default("retries", 3)
)
}
\arguments{
\item{indicator_code}{Character vector of 19-digit indicator codes. Names
become aliases. Requests are automatically batched in groups of five.}

\item{region_code}{Optional vector of five-digit Japanese region codes or
three-letter ISO country codes. Requests are automatically batched in
groups of 50.}

\item{parent_region_code}{Optional five-digit parent-region code.}

\item{region_level}{Optional official region-level code.}

\item{time, time_from, time_to}{Optional API time codes. Use \code{"20240100"} for
January 2024, \code{"20241Q00"} for 2024 Q1, \code{"2024CY00"} for calendar year
2024, or \code{"2024FY00"} for fiscal year 2024.}

\item{cycle}{Data cycle: \code{"month"}, \code{"quarter"}, \code{"year"}, or
\code{"fiscal_year"}; API codes 1--4 are also accepted.}

\item{regional_rank}{One of \code{"country"}, \code{"japan"}, \code{"prefecture"}, or
\code{"municipality"}; API codes 1--4 are also accepted.}

\item{seasonal}{\code{"original"} or \code{"seasonally_adjusted"}; API codes 1 and 2
are also accepted.}

\item{stat_name}{Optional partial match for a statistical survey name.}

\item{value_condition}{Optional strict numeric filter such as \code{">100"} or
\code{"<0"}.}

\item{wide}{If \code{FALSE} (default), return normalized long data. If \code{TRUE},
put each indicator or alias in its own column.}

\item{lang}{Response language, \code{"en"} or \code{"jp"}.}

\item{wait}{Requested seconds between automatically batched requests.
Values below one are treated as one when another request is required.}

\item{timeout}{Request timeout in seconds.}

\item{retries}{Number of retries for transient failures.}
}
\value{
A tibble. Long output retains raw codes and labels, an R \code{Date},
numeric \code{value}, original \code{value_raw}, and annotations.
}
\description{
\code{dashboard_data()} retrieves one or more indicators and returns one
indicator-region-time observation per row. The API permits five indicators
and 50 regions per request. Longer vectors are split automatically, with at
least one second between requests.
}
\details{
Named \code{indicator_code} vectors create aliases in the \code{indicator} column and
in wide output, following the style of \code{WDI::WDI()}.
}
\examples{
\dontrun{
population <- dashboard_data(
  indicator_code = c(population = "0201010000000010000"),
  region_code = "00000",
  time_from = "2020CY00",
  time_to = "2024CY00",
  cycle = "year",
  regional_rank = "japan",
  seasonal = "original"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credit.R
\name{dashboard_api_credit}
\alias{dashboard_api_credit}
\title{Return Statistics Dashboard attribution and API credit}
\usage{
dashboard_api_credit(lang = c("en", "jp"))
}
\arguments{
\item{lang}{Credit language, \code{"en"} or \code{"jp"}.}
}
\value{
A named list containing \code{source} (the required source citation),
\code{processed} (a template for edited content), \code{credit} (the required API
service credit), and official URLs.
}
\description{
The Statistics Dashboard asks users of its content to cite the source. When
publishing edited content, identify the editing entity as well. Publishers of
services using the API must also show a specified credit. The terms can
change, so verify the linked official page before publication.
}
\examples{
dashboard_api_credit("en")
dashboard_api_credit("jp")$credit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dashboardapi-package.R
\docType{package}
\name{dashboardapi-package}
\alias{dashboardapi}
\alias{dashboardapi-package}
\title{dashboardapi: Japan's Statistics Dashboard data for R}
\description{
\code{dashboardapi} is an unofficial, independently developed client for the
Statistics Dashboard Web API provided by the Statistics Bureau of Japan.
It is not affiliated with, endorsed by, or maintained by the Statistics
Bureau. A typical workflow is:
}
\details{
\enumerate{
\item Find a series with \code{\link[=dashboard_search]{dashboard_search()}}.
\item Find region codes with \code{\link[=dashboard_regions]{dashboard_regions()}}.
\item Retrieve observations with \code{\link[=dashboard_data]{dashboard_data()}}.
}

No registration or API key is required. The API documentation asks clients
not to create a large volume of access in a short period. When
\code{\link[=dashboard_data]{dashboard_data()}} must split a request into multiple API calls, the package
therefore waits at least one second between calls.
}
\section{Package options}{

\itemize{
\item \code{dashboardapi.lang}: default response language, \code{"en"} or \code{"jp"}.
\item \code{dashboardapi.wait}: seconds between automatically batched requests;
default and enforced minimum \code{1}.
\item \code{dashboardapi.timeout}: request timeout in seconds; default \code{30}.
\item \code{dashboardapi.retries}: retries for transient failures; default \code{3}.
}
}

\section{Terms of use}{

When publishing content retrieved with this package, cite the Statistics
Dashboard as the source. When publishing edited content, also identify the
editing entity. Services published using the API should show the credit
requested by the Statistics Dashboard. Call \code{\link[=dashboard_api_credit]{dashboard_api_credit()}} for
the source citation, edited-content template, API credit, and official
links. The package's MIT license applies only to original package code and
documentation; it does not relicense government or third-party data,
metadata, or official texts. Check any individual rights statement before
reusing third-party content.
}

\seealso{
\href{https://dashboard.e-stat.go.jp/en/static/api}{Statistics Dashboard API}
}
\author{
\strong{Maintainer}: Kenji Miyazaki \email{kenjimyzk@gmail.com}

Authors:
\itemize{
  \item Kenji Miyazaki \email{kenjimyzk@gmail.com}
}

}
\keyword{internal}
